"""
@author: Maxim Vavilin maxim.vavilin@kit.edu
"""

import numpy as np
import matplotlib.pyplot as plt

import repscat as rs

import fig_interaction_cross_section_of_speed
import fig_absorption_of_speed


def save_figures(file_name, y_val, y_label, xi_list, cross_sec_args, absorption_or_crossec):
    fig, ax = plt.subplots()
    linecolor = "darkred"
    ax.plot(np.tanh(xi_list), y_val, linestyle="-", color=linecolor)

    ax.set_xlabel("v/c")
    ax.set_ylabel(y_label, color=linecolor)
    ax.tick_params(axis="y", labelcolor=linecolor)

    ax.set_zorder(1)  # puts first plot on top of the second
    ax.patch.set_visible(False)

    if absorption_or_crossec == 'crossec':
        fig_interaction_cross_section_of_speed.plot(
            *cross_sec_args,
            ax.twinx(),
        )
    elif absorption_or_crossec == 'absorption':
        fig_absorption_of_speed.plot(*cross_sec_args, ax.twinx())

    fig.savefig(file_name, bbox_inches="tight")


def plot(incident_pw, jay_max, radius, xi_max, len_xi, absorption_or_crossec):
    momentum_transfer_list_lab = []
    energy_transfer_list_lab = []
    # xi_list = np.linspace(-0.005, 0.02, num_xi, endpoint=True)
    xi_list = np.linspace(-xi_max, xi_max, len_xi, endpoint=True)

    energy_transfer_list_obj = []
    momentum_transfer_list_obj = []

    for xi in xi_list:
        print("xi = ", xi)

        # Fields and Tmat in Object frame
        inc_pw_obj = incident_pw.boost(-xi)
        inc_am_obj = inc_pw_obj.in_angular_momentum_basis(jay_max)

        tmat_diag_obj = rs.get_silicon_tmat(radius, jay_max, inc_am_obj.k_list)

        scat_am_obj = tmat_diag_obj.scatter(inc_am_obj)

        scat_pw_obj = scat_am_obj.in_plane_wave_basis(
            inc_pw_obj.theta_list, inc_pw_obj.phi_list
        )

        # Transfer in object frame
        out_pw_obj = inc_pw_obj + scat_pw_obj

        mom_diff_obj = inc_pw_obj.momentum_z() - out_pw_obj.momentum_z()
        energy_diff_obj = inc_pw_obj.energy() - out_pw_obj.energy()

        energy_transfer_list_obj.append(energy_diff_obj)
        momentum_transfer_list_obj.append(mom_diff_obj)

        # Transfer in lab frame
        energy_diff_lab = (
            np.cosh(xi) * energy_diff_obj + np.sinh(-xi) * mom_diff_obj * rs.C_0_SI
        )
        mom_diff_lab = (
            np.sinh(-xi) * energy_diff_obj / rs.C_0_SI + np.cosh(xi) * mom_diff_obj
        )

        energy_transfer_list_lab.append(energy_diff_lab)
        momentum_transfer_list_lab.append(mom_diff_lab)

    file_names = (
        f"figs/Momentum_transfer_OBJ_frame_{len(xi_list)}_r_{radius}_{absorption_or_crossec}.png",
        f"figs/Energy_transfer_OBJ_frame_{len(xi_list)}_r_{radius}_{absorption_or_crossec}.png",
        f"figs/Momentum_transfer_LAB_frame_{len(xi_list)}_r_{radius}_{absorption_or_crossec}.png",
        f"figs/Energy_transfer_LAB_frame_{len(xi_list)}_r_{radius}_{absorption_or_crossec}.png",
    )

    y_vals = (
        np.real(momentum_transfer_list_obj),
        1e3 * np.real(energy_transfer_list_obj),
        np.real(momentum_transfer_list_lab),
        1e3 * np.real(energy_transfer_list_lab),
    )

    y_labels = (
        "Momentum in z transfer, $\\si{\kilogram\metre\per\second}$",
        "Energy transfer, $\\si{\milli\joule}$",
        "Momentum in z transfer, $\\si{\kilogram\metre\per\second}$",
        "Energy transfer, $\\si{\milli\joule}$",
    )
    cross_sec_args = (
        incident_pw.info["central_wavelength"],
        xi_max,
        len_xi,
        radius,
        jay_max,
    )

    for i in range(len(y_labels)):
        save_figures(file_names[i], y_vals[i], y_labels[i], xi_list, cross_sec_args, absorption_or_crossec)

    